@@ -4,17 +4,17 @@ import json
 import logging
 import os
 import re
+import sys
 import subprocess
 
 from jira.resources import Issue
 import six
 
 from . import constants
+from . import exceptions
 from . import migrations
-from .exceptions import (
-    CannotInferTicketNumberFromFolderName,
-    NotTicketFolderException
-)
+from .decorators import stash_local_changes
+from .jirafieldmanager import JiraFieldManager
 
 
 logger = logging.getLogger(__name__)
@@ -28,7 +28,7 @@ class TicketFolder(object):
         self.get_jira = jira
 
         if not os.path.isdir(self.metadata_dir):
-            raise NotTicketFolderException(
+            raise exceptions.NotTicketFolderException(
                 "%s is not a synchronizable ticket folder" % (
                     path
                 )
@@ -56,13 +56,21 @@ class TicketFolder(object):
             self._issue = self.jira.issue(self.ticket_number)
         return self._issue
 
-    def store_cached_issue(self):
+    def store_cached_issue(self, shadow=True):
         storable = {
             'options': self.issue._options,
             'raw': self.issue.raw
         }
-        with open(self.get_metadata_path('issue.json'), 'w') as out:
-            out.write(json.dumps(storable))
+        with open(
+            self.get_path('.jirafs/issue.json', shadow=shadow), 'w'
+        ) as out:
+            out.write(
+                json.dumps(
+                    storable,
+                    indent=4,
+                    sort_keys=True,
+                )
+            )
 
     @property
     def cached_issue(self):
@@ -100,7 +108,7 @@ class TicketFolder(object):
     def infer_ticket_number(self):
         raw_number = self.path.split('/')[-1:][0].upper()
         if not re.match('^\w+-\d+$', raw_number):
-            raise CannotInferTicketNumberFromFolderName(
+            raise exceptions.CannotInferTicketNumberFromFolderName(
                 "Cannot infer ticket number from folder %s. Please name "
                 "ticket folders after the ticket they represent." % (
                     self.path,
@@ -114,19 +122,29 @@ class TicketFolder(object):
             filename
         )
 
-    def get_remote_file_metadata(self):
-        remote_files = self.get_shadow_path('.jirafs/remote_files.json')
+    def get_remote_file_metadata(self, shadow=True):
+        remote_files = self.get_path(
+            '.jirafs/remote_files.json',
+            shadow=shadow
+        )
         try:
             with open(remote_files, 'r') as _in:
                 return json.loads(_in.read())
         except IOError:
             return {}
 
-    def set_remote_file_metadata(self, data):
-        remote_files = self.get_shadow_path('.jirafs/remote_files.json')
+    def set_remote_file_metadata(self, data, shadow=True):
+        remote_files = self.get_path(
+            '.jirafs/remote_files.json',
+            shadow=shadow
+        )
         with open(remote_files, 'w') as out:
             out.write(
-                json.dumps(data)
+                json.dumps(
+                    data,
+                    indent=4,
+                    sort_keys=True,
+                )
             )
 
     def get_local_path(self, filename):
@@ -141,6 +159,11 @@ class TicketFolder(object):
             filename,
         )
 
+    def get_path(self, filename, shadow=False):
+        if shadow:
+            return self.get_shadow_path(filename)
+        return self.get_local_path(filename)
+
     @property
     def version(self):
         try:
@@ -170,15 +193,18 @@ class TicketFolder(object):
                 '%s\n' % constants.METADATA_DIR,
             )
 
-        subprocess.check_call((
-            'git',
-            '--bare',
-            'init',
-            os.path.join(
-                metadata_path,
+        subprocess.check_call(
+            (
                 'git',
-            )
-        ))
+                '--bare',
+                'init',
+                os.path.join(
+                    metadata_path,
+                    'git',
+                )
+            ),
+            stdout=subprocess.PIPE
+        )
         subprocess.check_call((
             'git',
             'config',
@@ -199,7 +225,7 @@ class TicketFolder(object):
         instance.run_git_command(
             'commit', '--allow-empty', '-m', 'Initialized'
         )
-        instance.run_migrations(silent=True)
+        instance.run_migrations(init=True)
 
         comment_path = instance.get_local_path(constants.TICKET_NEW_COMMENT)
         with open(comment_path, 'w') as out:
@@ -212,12 +238,13 @@ class TicketFolder(object):
         path = os.path.realpath(path)
         os.mkdir(path)
         folder = cls.initialize_ticket_folder(path, jira)
-        folder.sync()
+        folder.pull()
         return folder
 
     def run_git_command(self, *args, **kwargs):
         failure_ok = kwargs.get('failure_ok', False)
         shadow = kwargs.get('shadow', False)
+        binary = kwargs.get('binary', False)
 
         if not shadow:
             work_tree = self.path,
@@ -235,21 +262,30 @@ class TicketFolder(object):
 
         self.log('Executing git command %s', (cmd, ), logging.DEBUG)
         try:
-            return subprocess.check_output(
+            result = subprocess.check_output(
                 cmd,
                 stderr=subprocess.PIPE
-            ).decode('utf-8').strip()
-        except subprocess.CalledProcessError:
+            )
+            if not binary:
+                return result.decode('utf-8').strip()
+            return result
+        except subprocess.CalledProcessError as e:
             if not failure_ok:
-                raise
+                raise exceptions.GitCommandError(
+                    "Error running command `%s`" % ' '.join(cmd),
+                    inner_exception=e
+                )
 
-    def get_local_file_at_revision(self, path, revision, failure_ok=True):
+    def get_local_file_at_revision(
+        self, path, revision, failure_ok=True, binary=False
+    ):
         return self.run_git_command(
             'show', '%s:%s' % (
                 revision,
                 path,
             ),
-            failure_ok=failure_ok
+            failure_ok=failure_ok,
+            binary=binary,
         )
 
     def get_ignore_globs(self, which=constants.IGNORE_FILE):
@@ -295,9 +331,22 @@ class TicketFolder(object):
                 return True
         return False
 
-    def get_locally_changed(self):
-        ignore_globs = self.get_ignore_globs()
+    def get_ready_changes(self):
+        changed_files = self.filter_ignored_files(
+            self.run_git_command(
+                'diff', '--name-only', self.git_merge_base,
+            ).split('\n')
+        )
 
+        return {
+            'fields': (
+                self.get_fields('HEAD') - self.get_fields(self.git_merge_base)
+            ),
+            'files': changed_files,
+            'new_comment': self.get_new_comment(ready=True)
+        }
+
+    def get_uncommitted_changes(self):
         new_files = self.run_git_command(
             'ls-files', '-o', failure_ok=True
         ).split('\n')
@@ -305,163 +354,98 @@ class TicketFolder(object):
             'ls-files', '-m', failure_ok=True
         ).split('\n')
 
-        all_possible = [
-            filename for filename in new_files + modified_files if filename
-        ]
+        return {
+            'files': self.filter_ignored_files([
+                filename for filename in new_files + modified_files if filename
+            ]),
+            'fields': self.get_fields() - self.get_fields('HEAD'),
+            'new_comment': self.get_new_comment(ready=False)
+        }
+
+    def get_remotely_changed(self):
+        metadata = self.get_remote_file_metadata(shadow=True)
 
         assets = []
-        for filename in all_possible:
+        attachments = self.filter_ignored_files(
+            self.issue.fields.attachment,
+            constants.REMOTE_IGNORE_FILE
+        )
+        for attachment in attachments:
+            changed = metadata.get(attachment.filename) != attachment.created
+            if changed:
+                assets.append(attachment.filename)
+
+        return assets
+
+    def filter_ignored_files(self, files, which=constants.IGNORE_FILE):
+        ignore_globs = self.get_ignore_globs(which)
+
+        assets = []
+        for fileish in files:
+            # Get the actual filename; this is a little gross -- apologies.
+            filename = fileish
+            attachment = False
+            if not isinstance(fileish, six.string_types):
+                filename = fileish.filename
+                attachment = True
+
             if self.file_matches_globs(filename, ignore_globs):
                 continue
-            if not os.path.isfile(os.path.join(self.path, filename)):
+            if (
+                not attachment
+                and not os.path.isfile(os.path.join(self.path, filename))
+            ):
                 continue
             if filename.startswith('.'):
                 continue
-            assets.append(filename)
+            assets.append(fileish)
 
         return assets
 
-    def get_remotely_changed(self):
-        ignore_globs = self.get_ignore_globs(constants.REMOTE_IGNORE_FILE)
-        metadata = self.get_remote_file_metadata()
-
-        assets = []
-        for attachment in self.issue.fields.attachment:
-            matches_globs = (
-                self.file_matches_globs(attachment.filename, ignore_globs)
-            )
-            changed = metadata.get(attachment.filename) != attachment.created
-            if not matches_globs and changed:
-                assets.append(attachment.filename)
-        return assets
-
-    def get_field_data_from_string(self, string):
-        """ Gets field data from an incoming string.
-
-        Parses through the string using the following RST-derived
-        pattern::
-
-            0 | FIELD_NAME::
-            1 |
-            2 |     VALUE
-
-        """
-        FIELD_DECLARED = 0
-        PREAMBLE = 1
-        VALUE = 2
-
-        state = None
-
-        data = {}
-        field_name = ''
-        value = ''
-        if not string:
-            return data
-        lines = string.split('\n')
-        for idx, line in enumerate(lines):
-            line = line.replace('\t', '    ')
-            if state == FIELD_DECLARED and not line:
-                state = PREAMBLE
-            elif (
-                (state == VALUE or state is None)
-                and re.match('^(\w+)::', line)
-            ):
-                if value:
-                    data[field_name] = value.strip()
-                    value = ''
-                state = FIELD_DECLARED
-                field_name = re.match('^(\w+)::', line).group(1)
-                if not field_name:
-                    raise ValueError(
-                        "Syntax error on line %s" % idx
-                    )
-            elif (state == PREAMBLE or state == VALUE):
-                state = VALUE
-                value = value + '\n' + line[4:]  # Remove first indentation
-        if value:
-            data[field_name] = value.strip()
-
-        return data
-
-    def get_local_fields(self):
-        try:
-            with open(
-                self.get_local_path(constants.TICKET_DETAILS), 'r'
-            ) as current_status:
-                values = self.get_field_data_from_string(current_status.read())
-
-                for field in constants.FILE_FIELDS:
-                    file_field_path = self.get_local_path(
-                        constants.TICKET_FILE_FIELD_TEMPLATE
-                    ).format(field_name=field)
-                    if os.path.exists(file_field_path):
-                        with open(file_field_path, 'r') as file_field:
-                            values[field] = file_field.read().strip()
-
-                return values
-        except IOError:
-            pass
-
-        return {}
-
-    def get_original_values(self):
-        values = self.get_field_data_from_string(
-            self.get_local_file_at_revision(
-                constants.TICKET_DETAILS,
-                self.git_merge_base,
-            )
+    def get_fields(self, revision=None, path=None):
+        kwargs = {}
+        if not revision:
+            kwargs['path'] = path if path else self.path
+        else:
+            kwargs['revision'] = revision
+        return JiraFieldManager.create(
+            self,
+            **kwargs
         )
-        for field in constants.FILE_FIELDS:
-            file_field_path = (
-                constants.TICKET_FILE_FIELD_TEMPLATE.format(field_name=field)
-            )
-            try:
-                values[field] = self.get_local_file_at_revision(
-                    file_field_path,
-                    self.git_merge_base,
-                    failure_ok=False
-                ).strip()
-            except subprocess.CalledProcessError:
-                # This file did not exist in head
-                pass
 
-        return values
-
-    def get_local_differing_fields(self):
-        """ Get fields that differ between local and the last sync
-
-        .. warning::
-
-           Does not support setting fields that were not set originally
-           in a sync operation!
-
-        """
-        local_fields = self.get_local_fields()
-        original_values = self.get_original_values()
-
-        differing = {}
-        for k, v in original_values.items():
-            if local_fields[k] != v:
-                differing[k] = (v, local_fields[k], )
-
-        return differing
-
-    def get_new_comment(self, clear=False):
+    def get_new_comment(self, clear=False, staged=False, ready=True):
         try:
             with open(
                 self.get_local_path(constants.TICKET_NEW_COMMENT), 'r+'
             ) as c:
-                comment = c.read().strip()
-                if clear:
-                    c.seek(0)
+                local_contents = c.read().strip()
+            if ready:
+                contents = self.get_local_file_at_revision(
+                    constants.TICKET_NEW_COMMENT,
+                    'HEAD'
+                )
+                if contents:
+                    contents = contents.strip()
+                else:
+                    contents = ''
+            else:
+                contents = local_contents
+
+            if not ready and contents == self.get_new_comment(ready=True):
+                contents = ''
+
+            if contents == local_contents and clear:
+                with open(
+                    self.get_local_path(constants.TICKET_NEW_COMMENT), 'r+'
+                ) as c:
                     c.truncate()
 
-            return comment
+            return contents
         except IOError:
             return ''
 
     def fetch(self):
-        file_meta = self.get_remote_file_metadata()
+        file_meta = self.get_remote_file_metadata(shadow=True)
 
         for filename in self.get_remotely_changed():
             for attachment in self.issue.fields.attachment:
@@ -475,7 +459,7 @@ class TicketFolder(object):
                         file_meta[filename] = attachment.created
                         download.write(attachment.get())
 
-        self.set_remote_file_metadata(file_meta)
+        self.set_remote_file_metadata(file_meta, shadow=True)
 
         with open(self.get_shadow_path(constants.TICKET_DETAILS), 'w') as dets:
             for field in sorted(self.issue.raw['fields'].keys()):
@@ -508,10 +492,9 @@ class TicketFolder(object):
                     elif field in constants.NO_DETAIL_FIELDS:
                         continue
 
-                    dets.write('%s::\n\n' % field)
+                    dets.write('* %s:\n' % field)
                     for line in value.replace('\r\n', '\n').split('\n'):
                         dets.write('    %s\n' % line)
-                    dets.write('\n')
 
         comments_filename = self.get_shadow_path(constants.TICKET_COMMENTS)
         with open(comments_filename, 'w') as comm:
@@ -531,47 +514,76 @@ class TicketFolder(object):
         )
         self.run_git_command('push', 'origin', 'jira', shadow=True)
 
+    @stash_local_changes
     def merge(self):
-        self.run_git_command('stash', '--include-untracked', failure_ok=True)
         self.run_git_command('merge', 'jira')
-        self.run_git_command('stash', 'pop', failure_ok=True)
 
     def pull(self):
         self.fetch()
         self.merge()
 
+    def commit(self, message, *args):
+        self.run_git_command(
+            'add', '-A', failure_ok=False
+        )
+        self.run_git_command(
+            'commit', '-m', message, *args, failure_ok=False
+        )
+
+    def is_up_to_date(self):
+        jira_commit = self.run_git_command('rev-parse', 'jira')
+        master_commit = self.run_git_command('rev-parse', 'master')
+
+        try:
+            self.run_git_command(
+                'merge-base', '--is-ancestor', jira_commit, master_commit,
+            )
+        except exceptions.GitCommandError:
+            return False
+        return True
+
+    @stash_local_changes
     def push(self):
         status = self.status()
 
-        file_meta = self.get_remote_file_metadata()
+        if not self.is_up_to_date():
+            raise exceptions.LocalCopyOutOfDate()
 
-        for filename in status['to_upload']:
-            with open(self.get_local_path(filename), 'rb') as upload:
-                self.log(
-                    'Uploading file "%s"',
-                    (filename, ),
-                )
-                # Delete the existing issue if there is one
-                for attachment in self.issue.fields.attachment:
-                    if attachment.filename == filename:
-                        attachment.delete()
-                attachment = self.jira.add_attachment(
-                    self.ticket_number,
-                    upload
-                )
-                file_meta[filename] = attachment.created
+        file_meta = self.get_remote_file_metadata(shadow=False)
 
-        comment = self.get_new_comment(clear=True)
+        for filename in status['ready']['files']:
+            upload = six.BytesIO(
+                self.get_local_file_at_revision(
+                    filename,
+                    'HEAD',
+                    binary=True
+                )
+            )
+            self.log(
+                'Uploading file "%s"',
+                (filename, ),
+            )
+            # Delete the existing issue if there is one
+            for attachment in self.issue.fields.attachment:
+                if attachment.filename == filename:
+                    attachment.delete()
+            attachment = self.jira.add_attachment(
+                self.ticket_number,
+                upload,
+                filename=filename,
+            )
+            file_meta[filename] = attachment.created
+
+        self.set_remote_file_metadata(file_meta, shadow=False)
+
+        comment = self.get_new_comment(clear=True, ready=True)
         if comment:
             self.log('Adding comment "%s"' % comment)
             self.jira.add_comment(self.ticket_number, comment)
 
-        values = self.get_original_values()
-
         collected_updates = {}
-        for field, diff_values in status['local_differs'].items():
+        for field, diff_values in status['ready']['fields'].items():
             collected_updates[field] = diff_values[1]
-            values[field] = diff_values[1]
 
         if collected_updates:
             self.log(
@@ -581,7 +593,13 @@ class TicketFolder(object):
             self.issue.update(**collected_updates)
 
         # Commit local copy
-        self.run_git_command('add', '-A', failure_ok=True)
+        self.run_git_command('reset', '--soft', failure_ok=True)
+        self.run_git_command(
+            'add', '.jirafs/remote_files.json', failure_ok=True
+        )
+        self.run_git_command(
+            'add', constants.TICKET_NEW_COMMENT, failure_ok=True
+        )
         self.run_git_command(
             'commit', '-m', 'Pushed local changes', failure_ok=True
         )
@@ -589,8 +607,7 @@ class TicketFolder(object):
         # Commit changes to remote copy, too, so we record remote
         # file metadata.
         self.run_git_command('fetch', shadow=True)
-        self.run_git_command('merge', 'master')
-        self.set_remote_file_metadata(file_meta)
+        self.run_git_command('merge', 'origin/master', shadow=True)
         self.run_git_command('add', '-A', shadow=True)
         self.run_git_command(
             'commit', '-m', 'Pulled remote changes',
@@ -598,25 +615,41 @@ class TicketFolder(object):
         )
         self.run_git_command('push', 'origin', 'jira', shadow=True)
 
-    def sync(self):
-        self.pull()
-        self.push()
-
     def status(self):
-        locally_changed = self.get_locally_changed()
-
         status = {
-            'to_upload': locally_changed,
-            'local_differs': self.get_local_differing_fields(),
-            'new_comment': self.get_new_comment()
+            'uncommitted': self.get_uncommitted_changes(),
+            'ready': self.get_ready_changes(),
+            'up_to_date': self.is_up_to_date(),
         }
 
         return status
 
-    def run_migrations(self, silent=False):
+    @stash_local_changes
+    def run_migrations(self, init=False):
         loglevel = logging.INFO
-        if silent:
+        if init:
             loglevel = logging.DEBUG
+        else:
+            if self.version < constants.MINIMUM_REPO_VERSION:
+                version = self.version
+                # Find the latest version able to open this file.
+                while version < constants.CURRENT_REPO_VERSION:
+                    if version in constants.VERSION_CEILINGS:
+                        maximum_version = constants.VERSION_CEILINGS[version]
+                    # We might not have a mapping for this exact repo
+                    # version, but we may for one of the versions after it
+                    version += 1
+                print(
+                    "This ticket folder is too old for the version of "
+                    "Jirafs you are using.\nPlease either use Jirafs version "
+                    "%s for accessing this ticket folder, or delete and "
+                    "re-clone this ticket folder using `jirafs clone %s`." % (
+                        maximum_version,
+                        self.ticket_number
+                    )
+                )
+                sys.exit(1)
+
         while self.version < constants.CURRENT_REPO_VERSION:
             migrator = getattr(
                 migrations,
